# -*- coding: utf-8 -*-
import os
import json
import warnings


_CLIENT = None


def get_client():
    global _CLIENT
    if _CLIENT is None:
        init()

    return _CLIENT


def init(username=None, password=None, **kwargs):
    """
    :param str username: 用户名，license模式下为'license'
    :param str password: 用户密码，license模式下为下发的license
    :param str protocol：传输协议，默认HTTP
    :param str url: 数据服务的URL
    :param int connect_timeout: 连接建立超时时间,默认5秒
    :param int request_timeout: 数据传输超时时间，默认300秒
    :param str compressor: 数据传输过程中的压缩算法，默认不使用
    :param int pool_size: 连接池大小，默认10
    :param int page_size: 数据请求分页大小，默认100000
    :return:
    """
    user = username
    pwd = password

    try:
        env = json.loads(os.getenv("hs_udata_CONF"))
    except:
        env = {
            "USERNAME": username,
            "PASSWORD": password
        }
    if user is None:
        user = env.get("USERNAME")
        pwd = env.get("PASSWORD")

    if pwd is None:
        raise KeyError("请设置有效Token值")

    assert user
    assert pwd

    from hs_udata.utils.config import read_config, write_config
    # 获取系统配置
    try:
        conf = read_config()['system']
    except:
        # 读取配置失败，使用默认配置启动
        warnings.warn("读取配置失败，请检查hs_udata.utils.config.yml")
        raise

    url = kwargs.pop("url", None)
    if url is None:
        url = conf.get("url")
    else:
        conf["url"] = url

    assert url

    protocol = kwargs.pop("protocol", None)
    if protocol is None:
        protocol = conf.get("protocol", "HTTP")
    else:
        if protocol != "HTTP":
            warnings.warn("protocol 当前只支持HTTP，设置的{}将不会生效".format(protocol))
        else:
            conf["protocol"] = "HTTP"

    con_timeout = kwargs.pop("connect_timeout", None)
    if con_timeout is None:
        con_timeout = conf.get("connect_timeout", 5)
    else:
        if int(con_timeout) <= 0 or int(con_timeout) > 1000:
            warnings.warn("connect_timeout 有效范围为(0, 1000]，设置的{}将不会生效".format(con_timeout))
        else:
            conf["connect_timeout"] = con_timeout

    timeout = kwargs.pop("request_timeout", None)
    if timeout is None:
        timeout = conf.get("request_timeout", 300)
    else:
        if int(timeout) <= 0 or int(timeout) > 100000:
            warnings.warn("request_timeout 有效范围为(0, 100000]，设置的{}将不会生效".format(timeout))
        else:
            conf["request_timeout"] = timeout

    pool_size = kwargs.pop("pool_size", None)
    if pool_size is None:
        pool_size = conf.get("pool_size", 10)
    else:
        if int(pool_size) <= 0 or int(pool_size) > 100:
            warnings.warn("pool_size 有效范围为(0, 100]，设置的{}将不会生效".format(pool_size))
        else:
            conf["pool_size"] = pool_size

    compressor = kwargs.pop("compressor", None)
    if compressor is None:
        compressor = conf.get("compressor", None)
    else:
        conf["compressor"] = compressor

    page_size = kwargs.pop("page_size", None)
    if page_size is None:
        page_size = conf.get("page_size", 100000)
    else:
        if int(page_size) <= 0 or int(page_size) > 100000:
            warnings.warn("page_size 有效范围为(0, 100000]，设置的{}将不会生效".format(page_size))
        else:
            conf["page_size"] = page_size

    config = {
        "protocol": protocol,
        "auth": {
            "username": user,
            "password": pwd,
        },
        "request_timeout": timeout,
        "connect_timeout": con_timeout,
        "url": url,
        "pool_size": pool_size,
        "compressor": compressor,
        "page_size": page_size
    }
    if protocol == "HTTP":
        from hs_udata.utils.connection_pool import ConnectionPool
        global _CLIENT
        _CLIENT = ConnectionPool(config)
    else:
        raise RuntimeError("传输协议无效，got protocol[{}]".format(protocol))

    # 保存配置文件
    write_config({'system': conf})

    try:
        # 更新用户名和密码
        import operator
        if operator.ne(user, env.get("USERNAME")) or operator.ne(pwd, env.get("PASSWORD")):
            # 更新环境配置信息中的user和pwd
            cmd = r'setx hs_udata_CONF "{\"USERNAME\": \"%s\", \"PASSWORD\": \"%s\"}"' % (user, pwd)
            os.system(cmd)
    except:
        pass

def environ():
    """
    获取环境配置信息
    :return:
    """
    global _CLIENT
    if _CLIENT:
        return _CLIENT.config

    return {}





__all__ = [
    "init",
    "environ",
    "get_data"
]


def init(username=None, password=None, **kwargs):
    """
    :param str username: 用户名，license模式下为'license'
    :param str password: 用户密码，license模式下为下发的license
    :param str protocol：传输协议，默认HTTP
    :param str url: 数据服务的URL
    :param int connect_timeout: 连接建立超时时间,默认5秒
    :param int timeout: 数据传输超时时间，默认30秒
    :param str compressor: 数据传输过程中的压缩算法，默认不使用
    :param int max_pool_size: 连接池大小，默认10
    :param int page_size: 数据请求分页大小，默认100000
    :return:
    """
    from hs_udata.utils.client import init as cms_init

    cms_init(username, password, **kwargs)
    return


def environ():
    """
    获取环境配置信息
    :return:
    """
    from hs_udata.utils.client import environ as cms_environ
    return cms_environ()


def get_data(method, **kwargs):
    """
    共用获取数据接口
    :param method:    接口名称
    :param kwargs:    接口参数
    """
    from hs_udata.utils.client import get_client

    return get_client().send(method, **kwargs)



# -*- coding: utf-8 -*-
import time
import inspect
from functools import wraps


def retry(count, exp_name, time_delta=1.0):
    def decorate(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            assert count
            exec_count = 1
            while exec_count < count:
                try:
                    return func(*args, **kwargs)
                except exp_name as ex:       #返回func时异常
                    exec_count += 1
                    if exec_count == count:
                        raise ex
                    print("Retry because %s", ex)
                    if time_delta:
                        time.sleep(time_delta)

        return wrap

    return decorate


functions = []


def lru_cache(*args, **kwargs):
    from functools import lru_cache as _lru_cache

    def decorator(func):
        func = _lru_cache(*args, **kwargs)(func)    #用_lru_cache进行缓存
        functions.append(func)      #将函数及其参数寸在列表中
        return func

    return decorator


def args_check(*checks):      #装饰器函数返回其子函数decorator
    def decorator(func):      #函数执行时再调用内层子函数
        @wraps(func)
        def wrap(*args, **kwargs):    #函数的参数传给wrap
            call_args = None
            for check in checks:          #逐步执行args_check中每一个参数
                if call_args is None:
                    call_args = inspect.getcallargs(get_original_func(func), *args, **kwargs)  #使用inspect模块的getcallargs方法返回一个将参数名字和值作为键值对的字典(按照形参和实际传入参数的位置形成键值对)

                check.check(func.__name__, call_args.get(check.arg_name, check.arg_name), kwargs)

            return func(*args, **kwargs)
        return wrap

    return decorator


def get_original_func(func):
    func_wrapped = getattr(func, "__wrapped__", None)          #获取函数__wrapped__属性，默认None
    while func_wrapped:                                        #通过循环获取最内层函数属性并最终返回
        func = func_wrapped
        func_wrapped = getattr(func, "__wrapped__", None)

    return func


class ArgumentChecker(object):

    def __init__(self, args_name):
        self._args_name = args_name
        self._functions = []
        self._required_func = None

    def is_instance(self, _types):
        from hs_udata.utils.error import InvalidArgument

        def warp(func, value):
            if not isinstance(value, _types):      #检查值是否是某一特定类型
                raise InvalidArgument("函数[{}] 参数类型异常，参数[{}]有效类型为[{}]，实参类型为[{}]".format(
                    func, self._args_name, _types, type(value)))

        self._functions.append(warp)          #调外层函数时将函数及其值存入，等内层函数执行时判断
        return self

    def is_required(self):
        from hs_udata.utils.error import InvalidArgument

        assert self.arg_name
        if not isinstance(self.arg_name, (tuple, list)):
            raise InvalidArgument("必填项装饰器，参数的有效类型为元组或数组，实参类型为[{}]".format(type(self._args_name)))

        def warp(func, value):
            checkd = False
            for name in self.arg_name:
                if name in value.keys() and value[name]:
                    checkd = True
                    break

            if not checkd:
                raise InvalidArgument("函数[{}] 必填项参数异常，参数[{}]至少填其中一个，实参传入为{}".format(
                    func, self._args_name, list(value.keys())))

        self._required_func = warp
        return self

    def check(self, func, value, kwargs):            #取出_functions内容进行检查
        if self._required_func:
            self._required_func(func, kwargs)
            return

        for check_func in self._functions:
            check_func(func, value)

    @property
    def arg_name(self):
        return self._args_name


def check(arg_name):      #在调用此函数时再调用ArgumentChecker类
    return ArgumentChecker(arg_name)






class ConnectionException(Exception):
    pass


class ConnectionTimeOut(Exception):
    pass


class RequestTimeOut(Exception):
    pass


class RequestException(Exception):
    pass


class ServerException(Exception):
    pass


class GatewayException(Exception):
    pass


class ArgumentsVaildError(Exception):
    pass


class InvalidArgument(Exception):
    pass


ERROR_CHANNEL_BASE = 9100
GATEWAY_ERROR_DICT = {
    ERROR_CHANNEL_BASE + 1: "no timestamp field",
    ERROR_CHANNEL_BASE + 2: "error timestamp format",
    ERROR_CHANNEL_BASE + 3: "error timestamp",
    ERROR_CHANNEL_BASE + 4: "no signature head",
    ERROR_CHANNEL_BASE + 5: "unkown HTTP method",
    ERROR_CHANNEL_BASE + 6: "resty_sm3:new() failed",
    ERROR_CHANNEL_BASE + 7: "sm3:update() failed",
    ERROR_CHANNEL_BASE + 8: "error packet",
    ERROR_CHANNEL_BASE + 9: "no appkey head",
    ERROR_CHANNEL_BASE + 10: "redis not config, please check...",
    ERROR_CHANNEL_BASE + 11: "no getinfo_byself.lua or no get_secret_byself function in getinfo_byself.lua",
    ERROR_CHANNEL_BASE + 12: "no app_secret in redis or hget from redis timeout",
    ERROR_CHANNEL_BASE + 13: "authentication failure,no client_key in clients",
    ERROR_CHANNEL_BASE + 14: "authentication failure,ip not in whitelist",
    ERROR_CHANNEL_BASE + 15: "authentication failure,ip in blacklist",
    ERROR_CHANNEL_BASE + 16: "authentication failure,app_auth_type is nil",
    ERROR_CHANNEL_BASE + 17: "no client_id in body",
    ERROR_CHANNEL_BASE + 18: "no client_id in args",
    ERROR_CHANNEL_BASE + 19: "error client id",
    ERROR_CHANNEL_BASE + 20: "no client_id head",
    ERROR_CHANNEL_BASE + 21: "error packet,no data_value with get request",
    ERROR_CHANNEL_BASE + 22: "must json type",
    ERROR_CHANNEL_BASE + 23: "error packet,not data_value in post_args",
    ERROR_CHANNEL_BASE + 24: "error packet,not data_value in header or args"
}

ERROR_SERVER_BASE = 100000
SERVER_ERROR_DICT = {
    ERROR_SERVER_BASE + 0: "发生异常",
    ERROR_SERVER_BASE + 1: "参数[{0}]解析失败!",
    ERROR_SERVER_BASE + 2: "请求方式不符,该api为{0}请求",
    ERROR_SERVER_BASE + 3: "参数[{0}]类型转换失败",
    ERROR_SERVER_BASE + 4: "数据库连接失败[{0}]",
    ERROR_SERVER_BASE + 5: "该语句不包含select关键字!",
    ERROR_SERVER_BASE + 6: "执行模式错误!",
    ERROR_SERVER_BASE + 7: "该数据库类型不支持!",
    ERROR_SERVER_BASE + 8: "查询失败:[{0}]",
    ERROR_SERVER_BASE + 9: "该引擎不支持!",
    ERROR_SERVER_BASE + 10: "查询失败",
    ERROR_SERVER_BASE + 11: "参数[{0}]不能为空",
    ERROR_SERVER_BASE + 12: "驱动找不到,{0}",
    ERROR_SERVER_BASE + 13: "该数据库不支持统一分页,请自行使用sql语句分页",
    ERROR_SERVER_BASE + 14: "数据推送失败{0}",
    ERROR_SERVER_BASE + 15: "暂不支持get请求",
    ERROR_SERVER_BASE + 16: "csv文件写入失败",
    ERROR_SERVER_BASE + 17: "ftp文件上传失败",
    ERROR_SERVER_BASE + 18: "暂不支持该输出通道",
    ERROR_SERVER_BASE + 19: "kerberos认证失败[{0}]",
    ERROR_SERVER_BASE + 20: "kerberos认证失败,文件[{0}]不存在",
    ERROR_SERVER_BASE + 24: "ftp文件上传失败,登录失败!",
    ERROR_SERVER_BASE + 25: "ftp连接失败!",
    ERROR_SERVER_BASE + 26: "通知发送失败!",
    ERROR_SERVER_BASE + 27: "请求格式错误",
    ERROR_SERVER_BASE + 28: "ftp创建文件夹[{}]失败!",
    ERROR_SERVER_BASE + 29: "数据源{0}下的数据表{1}不存在"
}




import os
import yaml

from hs_udata.utils.decorators import lru_cache


def load_yml(path):
    with open(path, encoding='utf-8') as pf:
        return yaml.safe_load(pf)


def save_yml(full_path, _yml):
    with open(full_path, encoding='utf-8', mode='w') as pf:
        return yaml.safe_dump(_yml, pf)


@lru_cache(128)
def read_config():
    path = os.path.join(os.path.dirname(__file__), ".", "config.yml")
    return load_yml(path)


def write_config(_config):
    conf = read_config()['system']
    import operator
    if operator.ne(_config, conf):
        path = os.path.join(os.path.dirname(__file__), ".", "config.yml")
        return save_yml(path, _config)



from enum import Enum

# 服务交互字段定义
ERROR_NO = "error_no"    # 错误代码域
ERROR_INFO = "error_info"  # 错误信息域
DATA = "data"    # 数据域


class ConnectionStatus(Enum):                 #连接代码
    Disconnected = 0x0000       # 未连接
    Connecting = 0x0001,        # 正在连接
    Connected = 0x0002,         # 已连接
    SafeConnecting = 0x0004,    # 正在建立安全连接
    SafeConnected = 0x0008,     # 已建立安全连接
    Registering = 0x0010,       # 正注册
    Registered = 0x0020,        # 已注册
    Rejected = 0x0040           # 被拒绝，将被关闭


class Protocol(Enum):                 #连接类型
    HTTP = "HttpConnection"
    TCP = "TCPConnection"



import os
from setuptools import setup,find_packages

path = os.path.abspath(os.path.dirname(__file__))

try:
  with open(os.path.join(path, 'README.md')) as f:
    long_description = f.read()
except Exception as e:
    long_description = "hs_udata from hundsun"

setup(
    name = "hs_udata",
    version = "0.2.4",
    long_description = long_description,
    python_requires=">=3.0.0",
    license = "MIT Licence",

    url = "https://udata.hs.net/home?channel_source=pypi",
    author = "恒生Light云",
    author_email = "huhl37814@hundsun.com",
    description=u'恒有数，源自恒生电子的金融数据社区，提供涵盖股票、基金、债券、期权期货、港股等金融数据；提供在线预览、在线下载和在线调试等功能，提供简单高效的API接口（接口语言包括HTTP、Python、MATLAB、Java）；提供丰富的接口文档、帮助文档与社区社群运营，使得用户可以方便快捷地取用数据。',
    install_requires = [
        "pandas>=0.25.2",
        "pyyaml>=5.3.1",
        "requests>=2.23.0"
    ],
    packages=find_packages(),
    data_file = "hs_udata\\utils\\config.yml",
    include_package_data = True,
    #包含所有.xlsx文件
    package_data = {
        '':['*.yml'],
    }


)


# -*- coding: utf-8 -*-
import datetime


def convert_date(dt, now=True):
    # if dt == "":    #时间为空时返回当前时间或19900101
    #     if now:
    #         return datetime.datetime.now().date().strftime("%Y%m%d")
    #     else:
    #         return "19900101""19900101"
    if dt == None:
        return

    if isinstance(dt, datetime.datetime):       #datetime类型数据进行数据转换成固定格式
        return dt.date().strftime("%Y%m%d")

    if isinstance(dt, datetime.date):          #datetime.date数据类型转换为固定格式
        return dt.strftime("%Y%m%d")

    if isinstance(dt, str):
        from dateutil.parser import parse
        return parse(dt).date().strftime("%Y%m%d")    #用parse进行格式转换，变成数据类型

    if hasattr(dt, "to_pydatetime"):                  #判断一个对象里面是否有"to_pydatetime"属性或者"to_pydatetime"方法
        return dt.to_pydatetime().date().strftime("%Y%m%d")

    if hasattr(dt, "dtype") and dt.dtype.char == "M":    #M表示带分钟的dtype类型
        from dateutil.parser import parse
        return parse(dt).date().strftime("%Y%m%d")

    raise ValueError("转换日期格式异常，实参为[{}]".format(dt))        #不在上述所列类型抛出异常


# -*- coding: utf-8 -*-
from collections import deque
from threading import Lock, Semaphore   #Semaphore 是用于控制进入数量的锁，控制同时进行的线程
from contextlib import contextmanager   #上下文管理

from hs_udata.utils.connection import Connection
from hs_udata.utils.decorators import retry


class ConnectionPool(object):
    def __init__(self, config):
        self._config = config.copy()
        self._lock = Lock()
        self._connections = deque()
        self._connect_timeout = config.pop("connect_timeout")
        self._signal = Semaphore(config.pop("pool_size"))

    @property
    def config(self):
        return self._config

    @retry(3, exp_name="RequestTimeOut")
    def send(self, method, **kwargs):
        with self._signal:
            with self._yield_conn() as conn:
                return conn.send(method, **kwargs)

    @contextmanager
    def _yield_conn(self):
        conn = self._get_connection()
        try:
            yield conn                #返回连接
        except Exception as ex:
            conn.close()              #连接异常时关闭连接
            raise ex
        else:
            with self._lock:
                self._connections.append(conn)         #若已加锁则将当前连接加入连接队列

    def _get_connection(self):
        with self._lock:
            while self._connections:
                conn = self._connections.popleft()        #从队列左边取出一个连接
                if conn.check():                          #检查连接是否可用
                    return conn
                else:
                    conn.close()

        return self._cretate_connection()

    @retry(5, exp_name="ConnectTimeout")
    def _cretate_connection(self):
        return Connection.get_connection(self._config)    #创建连接

    def close(self):
        with self._lock:
            for conn in self._connections:       #逐个关闭连接
                conn.close()
            self._connections.clear()            #清空队列



import warnings
import importlib
import pkgutil

__version__ = '0.0.3'


def _init():
    for loader, module_name, is_pkg in pkgutil.walk_packages(__path__, "hs_udata."):
        if module_name.startswith("hs_udata.apis") and not is_pkg:
            try:
                api_module = importlib.import_module(module_name)
            except ImportError as ex:
                warnings.warn("import module[{}] error, msg={}".format(module_name, ex))

            for api_name in api_module.__all__:
                try:
                    api = getattr(api_module, api_name)
                    globals()[api_name] = api
                except AttributeError as ex:
                    warnings.warn("load api[{}] error, msg={}".format(api_name, ex))

_init()

del _init

def set_token(token=None):
    from hs_udata.utils.client import init
    init(username="license", password=token)


# -*- coding: utf-8 -*-


def convert_param(_type, param):
    if param is None:
        return _type()              #返回该类型空值

    return param


def convert_fields(field):
    if field is None:
        return []

    if isinstance(field, list):
        return field

    if isinstance(field, str):
        return [field]

    raise ValueError("字段[{}]类型转换失败，支持[None, str, list]类型，实参类型为[{}]".format(field, type(field)))



# -*- coding: utf-8 -*-
import requests
import pandas as pd
import time
import six
import abc
import warnings
import os


from hs_udata.utils.common import (
    ConnectionStatus,
    ERROR_NO,
    ERROR_INFO,
    DATA
)
from hs_udata.utils.error import (
    ConnectionException,
    ConnectionTimeOut,
    RequestTimeOut,
    RequestException,
    GatewayException,
    ServerException,
    GATEWAY_ERROR_DICT,
    SERVER_ERROR_DICT,
)


class AbstractConnection(six.with_metaclass(abc.ABCMeta)):
    """
    连接抽象类
    """
    def create(self, *args, **kwargs):
        """
        创建连接对象
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def connect(self, timeout=5):
        """
        启动连接
        :param timeout:
        :return:
        """
        raise NotImplementedError

    def set_timeout(self, timeout):
        """
        设置超时时间
        :param timeout:
        :return:
        """
        raise NotImplementedError

    def send(self, *args, **kwargs):
        """
        同步发送数据
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def async_send(self, *args, **kwargs):
        """
        异步发送数据
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def receive(self, *args, **kwargs):
        """
        接收数据
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def status(self):
        """
        连接状态
        :return:
        """
        raise NotImplementedError

    def check(self):
        """
        检查连接是否可用
        :return:
        """
        raise NotImplementedError

    def error(self):
        """
        连接错误信息
        :return:
        """
        raise NotImplementedError

    def close(self):
        """
        关闭链接
        :return:
        """
        raise NotImplementedError


class Connection(object):
    """
    连接工厂类
    """
    @classmethod   #classmethod 修饰符对应的函数不需要实例化，不需要 self 参数，但第一个参数需要是表示自身类的 cls 参数，可以来调用类的属性，类的方法，实例化对象等。
    def get_connection(cls, config):
        from hs_udata.utils.common import Protocol
        protocol = config.get("protocol")
        conn_timeout = config.get("connect_timeout")
        try:
            with ConnectionContext(conn_timeout):
                return globals()[Protocol[protocol].value]().create(config)
        except KeyError:
            raise RuntimeError("创建链接对象失败,请检查配置信息是否正确？config={}".format(config))


class HttpConnection(AbstractConnection):
    """
    HTTP连接对象类
    """
    import requests

    def create(self, config):
        """
        创建连接对象
        :param gateway:
        :param auth:
        :return:
        """
        # 解析config文件中的信息
        url = config.get("url")
        auth = config.get("auth")

        assert url
        assert auth

        self._url = url
        self._user = auth.get("username")
        self._pwd = auth.get("password")
        self._timeout = config.get("request_timeout")
        self._page_size = config.get("page_size")

        assert self._user
        assert self._pwd

        self._conn = requests.Session()            # 创建一个session对象，requests库的session对象还能提供请求方法的缺省数据，通过设置session对象的属性来实现
        self._status = ConnectionStatus.Connected         #调common中的连接状态码，状态设为已连接
        return self

    def set_timeout(self, timeout):
        """
        设置超时时间
        :param timeout:
        :return:
        """
        self._timeout = timeout

    def send(self, method, **kwargs):
        """
        同步发送数据
        :param args:
        :param kwargs:
        :return:
        """
        if self.check():        #已连接状态时
            if self._user == "license":
                url_path = kwargs.get("url_path", "")
                url = "".join((self._url, "/", url_path, "/", method))
            else:
                raise RuntimeError("当前只支持license登录方式")
            header = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Application-Token": self._pwd
            }
            params = kwargs
            try:
                method_data = kwargs.get("method", "GET")
                if method_data.upper() == "POST":
                    response = self._conn.post(url, headers=header, params=params, timeout=self._timeout)
                else:
                    response = self._conn.get(url, headers=header, params=params, timeout=self._timeout)
            except requests.exceptions.ConnectTimeout as ex:
                raise RequestTimeOut(ex)
            if response.status_code == 200:   #状态码200表示请求成功
                rsp = response.json()         #requests获取后直接调用json()方法转变为python字典请求获取HttpResponse对象

                # 从数据服务返回的数据
                if "error_code" in rsp:
                    if int(rsp["error_code"]) == 0:
                        df = pd.DataFrame(rsp["data"]) if rsp.get("data") else pd.DataFrame()
                        df.columns = df.columns.map(lambda x:x.lower())
                        if kwargs.get("rslt_type", 0) == 0:
                            return df
                        else:
                            import numpy as np
                            return np.array(df.to_records())

                    if SERVER_ERROR_DICT.get(int(rsp["error_code"]), None):      #获取error中异常信息
                        # warnings.warn("app_key权限认证失败，请联系管理员，error_info={}".format(rsp["resultMsg"]))
                        # warnings.warn("app_key权限已过期，请重新申请，申请站点：{}".format(rsp["resultMsg"]))
                        warnings.warn("请求异常，异常信息：{}".format(rsp["result_msg"]))
                        raise RequestException(rsp["result_msg"])
                    else:
                        warnings.warn("获取数据异常，path={}，params={}，error_info={}".
                                      format(method, kwargs, rsp["result_msg"]))
                        raise RequestException("请求发生未定义错误，错误信息：{}".format(rsp["result_msg"]))
            elif response.status_code >= 400 and response.status_code < 500:
                rsp = response.json()
                # 从新版网关返回异常
                if ERROR_NO in rsp:
                    if GATEWAY_ERROR_DICT.get(rsp[ERROR_NO], None):      #获取错误代码
                        warnings.warn("网关返回错误，错误信息：{}".format(rsp[ERROR_INFO]))
                        raise GatewayException(rsp[ERROR_INFO])
                    else:
                        raise GatewayException("网关发生未定义错误，错误信息：{}".format(rsp[ERROR_INFO]))
                # 从老版网关返回异常
                elif DATA in rsp:
                    if isinstance(rsp[DATA], list):
                        error_dict = rsp[DATA][0]
                    else:
                        error_dict = rsp[DATA]

                    if GATEWAY_ERROR_DICT.get(error_dict[ERROR_NO], None):
                        warnings.warn("网关返回错误，错误信息：{}".format(error_dict[ERROR_INFO]))
                        raise GatewayException(error_dict[ERROR_INFO])
                    else:
                        raise GatewayException("网关发生未定义错误，错误信息：{}".format(error_dict[ERROR_INFO]))
                else:
                    warnings.warn("网关返回未定义异常，异常信息：{}".format(response.text))
                    raise GatewayException(response.text)

            elif response.status_code >=500 and response.status_code < 601:            #服务错误
                warnings.warn("服务异常，异常信息：{}".format(response.text))
                raise ServerException(response.text)
            else:
                warnings.warn("异常流程，返回信息：{}".format(response.text))          #其他流程错误
                raise Exception(response.text)

        raise ConnectionException("连接已关闭，无法获取数据")          #如果没连接则raise连接异常

    def async_send(self, *args, **kwargs):
        """
        异步发送数据
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError          #尚未实现的方法错误

    def receive(self, *args, **kwargs):
        """
        接收数据
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def status(self):
        """
        连接状态
        :return:
        """
        return self._status

    def check(self):
        return self._status in [ConnectionStatus.Connected, ConnectionStatus.SafeConnected]

    def error(self):
        """
        连接错误信息
        :return:
        """
        raise NotImplementedError

    def close(self):
        """
        关闭链接
        :return:
        """
        self._conn.close()
        self._status = ConnectionStatus.Disconnected


class ConnectionContext(object):     #自定义上下文管理器
    def __init__(self, timeout=5):
        assert timeout
        self.timeout = timeout
        self.start = None

    def __enter__(self):
        self.start = time.time()    #返回当前时间的时间戳（1970纪元后经过的浮点秒数）。

    def __exit__(self, exc_type, exc_val, exc_tb):
        if time.time() - self.start > self.timeout:     #当前花费时间超时
            raise ConnectionTimeOut
